/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.operator.keycloak.realm;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.keycloak.representations.idm.RealmRepresentation;
import org.keycloak.representations.userprofile.config.UPConfig;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.operator.keycloak.KeycloakGenericRemoteService;
import de.ozgcloud.operator.keycloak.realm.OzgCloudKeycloakRealmSpec.RealmRole;

class KeycloakRealmServiceTest {

	private static final OzgCloudKeycloakRealmSpec REALM = OzgCloudKeycloakRealmSpecTestFactory.create();
	private static final String REALM_NAME = "TestRealmName";
	private static final String ROLE_NAME = OzgCloudKeycloakRealmSpecTestFactory.ROLE_NAME_1;
	private static final RealmRole ROLE = OzgCloudKeycloakRealmSpecTestFactory.ROLE1;

	@Spy
	@InjectMocks
	private KeycloakRealmService service;

	@Mock
	private KeycloakRealmRemoteService remoteService;

	@Mock
	private KeycloakRealmMapper mapper;

	@Mock
	private RealmRepresentation realmRepresentation;

	@Mock
	private KeycloakGenericRemoteService keycloakGenericRemoteService;

	@Nested
	class TestCreateOrUpdateRealm {

		@Test
		void shouldCallCreateRealmMethodIfNotExists() {

			when(keycloakGenericRemoteService.getRealmRepresentation(REALM_NAME)).thenReturn(Optional.empty());

			service.createOrUpdateRealm(REALM, REALM_NAME);

			verify(service).createRealm(REALM, REALM_NAME);
		}

		@Test
		void shouldCallUpdateRealmIfAlreadyExists() {
			var existingRealm = RealmRepresentationTestFactory.create();
			when(keycloakGenericRemoteService.getRealmRepresentation(REALM_NAME)).thenReturn(Optional.of(existingRealm));
			doNothing().when(service).updateUserProfileConfig(any());

			service.createOrUpdateRealm(REALM, REALM_NAME);

			verify(service).updateRealm(existingRealm, REALM);
		}

		@Test
		void shouldCallGetRealmRepresentation() {
			service.createOrUpdateRealm(REALM, REALM_NAME);

			verify(keycloakGenericRemoteService).getRealmRepresentation(REALM_NAME);
		}

		@Test
		void shouldCallAddOrUpdateRealmRoles() {
			service.createOrUpdateRealm(REALM, REALM_NAME);

			verify(service).addOrUpdateRealmRoles(REALM, REALM_NAME);
		}
	}

	@DisplayName("Update Realm")
	@Nested
	class TestUpdateRealm {

		@BeforeEach
		void init() {
			when(mapper.update(any(), any())).thenReturn(realmRepresentation);
			when(remoteService.getUserProfileConfig(realmRepresentation)).thenReturn(new UPConfig());
		}

		@Test
		void shouldUpdateRealmIfExists() {
			updateRealm();

			verify(remoteService).updateRealm(realmRepresentation);
		}

		@Test
		void shouldCallMapper() {
			updateRealm();

			verify(mapper).update(realmRepresentation, REALM);
		}

		@Test
		void createRealmShouldCallAddAttributes() {
			updateRealm();

			verify(service).addUserProfileAttributes(realmRepresentation);
		}

		private void updateRealm() {
			service.updateRealm(realmRepresentation, REALM);
		}
	}

	@DisplayName("Add or Update Realm Roles")
	@Nested
	class TestAddOrUpdateRealmRoles {

		@Test
		void shouldCallRemoteServiceGetRole() {

			service.addOrUpdateRealmRoles(REALM, REALM_NAME);

			verify(remoteService).getRealmRole(ROLE_NAME, REALM_NAME);
		}

		@Test
		void shouldCreateRole() {
			when(remoteService.getRealmRole(ROLE_NAME, REALM_NAME)).thenReturn(Optional.empty());
			var roleRepresentation = RealmRoleRepresentationTestFactory.create();
			when(mapper.map(ROLE)).thenReturn(roleRepresentation);

			service.addOrUpdateRealmRoles(REALM, REALM_NAME);

			verify(remoteService).addRealmRole(roleRepresentation, REALM_NAME);
		}

		@Test
		void shouldUpdateRole() {
			var roleRepresentation = RealmRoleRepresentationTestFactory.create();
			when(mapper.map(ROLE)).thenReturn(roleRepresentation);
			when(remoteService.getRealmRole(ROLE_NAME, REALM_NAME)).thenReturn(Optional.of(roleRepresentation));

			service.addOrUpdateRealmRoles(REALM, REALM_NAME);

			verify(remoteService).updateRealmRole(roleRepresentation, REALM_NAME);
		}

	}

	@Nested
	class TestCreateRealm {

		@BeforeEach
		void init() {
			when(mapper.map(REALM)).thenReturn(realmRepresentation);
			when(remoteService.getUserProfileConfig(realmRepresentation)).thenReturn(new UPConfig());
		}

		@Test
		void shouldCallMapper() {
			service.createRealm(REALM, REALM_NAME);

			verify(mapper).map(REALM);
		}

		@Test
		void shouldCallCreateRealmIfNotExists() {
			service.createRealm(REALM, REALM_NAME);

			verify(remoteService).createRealm(realmRepresentation);
		}

		@Test
		void shouldCallAddRealmName() {
			service.createRealm(REALM, REALM_NAME);

			verify(service).addRealmName(realmRepresentation, REALM_NAME);
		}

		@Test
		void updateRealmShouldCallAddAttributes() {
			service.createRealm(REALM, REALM_NAME);

			verify(service).addUserProfileAttributes(realmRepresentation);
		}
	}

	@Test
	void testAddRealmName() {

		service.addRealmName(realmRepresentation, REALM_NAME);

		verify(realmRepresentation).setRealm(REALM_NAME);
	}

	@Nested
	class TestDeleteRealm {

		@Test
		void shouldCallRemoteServiceDelete() {
			service.deleteRealm(REALM_NAME);

			verify(remoteService).deleteRealm(REALM_NAME);
		}
	}

	@Nested
	class TestRealmExists {

		@Test
		void shouldCallGenericRemoteServiceRealmExists() {
			service.realmExists(REALM_NAME);

			verify(keycloakGenericRemoteService).realmExists(REALM_NAME);
		}
	}

	@Nested
	class TestAddUserProfileAttributes {

		private UPConfig userProfileConfig;

		@BeforeEach
		void init() {
			userProfileConfig = new UPConfig();
			when(remoteService.getUserProfileConfig(realmRepresentation)).thenReturn(userProfileConfig);
		}

		@Test
		void shouldCallRemoteServiceGetUserProfileConfig() {
			service.addUserProfileAttributes(realmRepresentation);

			verify(remoteService).getUserProfileConfig(realmRepresentation);
		}

		@Test
		void shouldCallRemoteServiceUpdateUserProfileConfig() {
			service.addUserProfileAttributes(realmRepresentation);

			verify(remoteService).updateUserProfileConfig(realmRepresentation, userProfileConfig);
		}

		@Test
		void shouldCallUpdateUserProfileConfig() {
			service.addUserProfileAttributes(realmRepresentation);

			verify(service).updateUserProfileConfig(any(UPConfig.class));
		}
	}

	@Nested
	class TestUpdateUserProfileConfig {

		@Test
		void shouldAddUserProfileAttribute() {
			var userProfileConfig = new UPConfig();

			service.updateUserProfileConfig(userProfileConfig);

			assertThat(userProfileConfig.getAttributes()).hasSize(1);
		}

		@Test
		void shouldAddUserProfileAttributeOzgCloud() {
			var userProfileConfig = new UPConfig();

			service.updateUserProfileConfig(userProfileConfig);

			var attributeName = userProfileConfig.getAttributes().get(0).getName();
			assertThat(attributeName).isEqualTo("ozgCloudUserId");
		}

		@Test
		void shouldSetPermissions() {
			var userProfileConfig = new UPConfig();

			service.updateUserProfileConfig(userProfileConfig);

			var permissions = userProfileConfig.getAttributes().get(0).getPermissions();
			assertThat(permissions.getEdit()).contains("admin", "user");
			assertThat(permissions.getView()).contains("admin", "user");
		}
	}
}
